using Nemerle.Collections;
using Nemerle.Internal;
using Nemerle.Compiler;
using Nemerle.Compiler.Parsetree;

namespace Oyster.Units.Macros
{
	/// Contains type generation methods
	module TypeGenerator
	{
		/// Contains already created unit literals
		mutable ProcessedLiterals : Set[string] = Set();

		/// Generates new unit type
		public GenerateUnitType(env : GlobalEnv, name : string, coef : double, powers : list[int]) : TypeBuilder
		{
			def nameIsEmpty = name == "";

			// Form type name
			def n = if (nameIsEmpty) Macros.NewSymbol("Unit") else Macros.UseSiteSymbol(name);

			// Try to find this type before creating it
			when (env.LookupType([ n.Id ]).HasValue)
			{
				Message.FatalError($"Type $(env.CurrentNamespace).$name already exists")
			};

			def ifaceName = Util.UseSiteExpr(GetUnitInterface(powers).FullName);
			def t = env.Define(<[ decl:
				[NormCoef($(coef : double))]
				public struct $(n : name)['a] : $ifaceName where 'a : $ifaceName
				{
					_value : double;

					public this(value : double)
					{
						_value = value
					}
					public this(unit : 'a)
					{
						_value = unit.Value * unit.NormCoef / $(coef : double)
					}

					public Value : double
					{
						get { _value }
					}
					public NormCoef : double
					{
						get { $(coef : double) }
					}

					override public ToString() : string
					{
						_value.ToString()
					}

					static public @:>(unit : $(n : name)['a]) : double
					{
						unit._value
					}
					static public @:(value : double) : $(n : name)['a]
					{
						$(n : name)(value)
					}
					
					// Not works because of the bug:
					// http://nemerle.org/bugs/view.php?id=630
					// 
					//static public @:['b](unit : 'b) : $(n : name)['a] where 'b : $ifaceName
					//{
					//	$(n : name)(unit)
					//}

					// Small workaround - will allow to pass units into functions without type usage
					// (using strong aliases).
					// Needs to be removed once previous implicit op will be uncommented
					static public @:['b](unit : $(n : name)['b]) : $(n : name)['a] where 'b : $ifaceName
					{
						$(n : name)(unit.Value)
					}

					static public @++(unit : $(n : name)['a]) : $(n : name)['a]
					{
						$(n : name)(unit.Value + 1)
					}
					static public @--(unit : $(n : name)['a]) : $(n : name)['a]
					{
						$(n : name)(unit.Value - 1)
					}
					static public @+(unit : $(n : name)['a]) : $(n : name)['a]
					{
						unit
					}
					static public @-(unit : $(n : name)['a]) : $(n : name)['a]
					{
						$(n : name)(0 - unit.Value)
					}
				}
			]>);

			// Define all simple arithmetic operations
			def defineSimpleOp(opName, valueType, opFun, isLeft)
			{
				mutable args = [ <[ parameter: value : $valueType ]>, <[ parameter: unit : $(n : name)['a] ]> ];
				mutable opArgs = [ <[ value ]>, <[ unit.Value ]> ];
				when (!isLeft)
				{
					args = args.Reverse();
					opArgs = opArgs.Reverse();
				};
				
				t.Define(<[ decl:
					static public $(opName.name : name)( .. $args ) : $(n : name)['a]
					{
						$(n : name)($(opFun(opArgs.Head, opArgs.Tail.Head)))
					}
				]>)
			};
			def defineOp(opName, opFun)
			{
				[ true, false ].Iter(defineSimpleOp(opName, <[ double ]>, opFun, _))
			};

			defineOp(<[ @+ ]>, Util.AddExpr);
			defineOp(<[ @- ]>, Util.SubExpr);
			defineOp(<[ @* ]>, Util.MulExpr);
			defineOp(<[ @/ ]>, Util.DivExpr);
			
			t.Compile();

			// Also define strong aliases - for convenience
			def ta = env.Define(<[ decl:
				public type $(n : name) = $(n : name)[$ifaceName];
			]>);
			ta.Compile();

			t
		}

		/// Generates or gets unit interface for given powers list
		GetUnitInterface(powers : list[int]) : TypeInfo
		{
			def env = Util.UnitsEnv;

			def ifaceName = $"$(Util.UnitInterfacePrefix)$(Util.GetUnitSuffix(powers))";
			match (env.LookupType([ ifaceName ]))
			{
				| Some(t) => t
				| _ =>
					// Create new interface
					def ifaceName = Macros.UseSiteSymbol(ifaceName);
					def t = env.Define(<[ decl:
						public interface $(ifaceName : name) : IUnit {}
					]>);
					t.Compile();
					t
			}
		}

		/// Generates unit literal macro (actually macro-operator)
		public GenerateUnitLiteralMacro(opName : string, unitFullName : string) : void
		{
			// Check if given literal was already created
			when (ProcessedLiterals.Contains(opName))
			{
				Message.FatalError($"$opName unit literal is duplicated in declaration")
			};

			// Create new macro
			def n = opName + "_UnitLiteralMacro";
			def t = Util.MacrosEnv.Define(<[ decl:
				public class $(n : usesite) : UnitLiteralMacro
				{
					public this()
					{
						base($(opName : string), $(unitFullName : string))
					}
				}
			]>);
			t.Compile();

			// Emit assembly attributes for macro
			def asmAttr(expr) { AttributeCompiler.AddAssemblyAttribute(Util.MacrosEnv, expr) };
			asmAttr(<[ ContainsMacro($($"$(Util.MacrosNS).$n" : string)) ]>);
			asmAttr(<[ Operator($(Util.MacrosNS : string), $(opName : string), true, 283, 284) ]>);

			ProcessedLiterals = ProcessedLiterals.Add(opName)
		}
	}
}
