using System;
using Nemerle.Compiler;
using Nemerle.Compiler.Parsetree;
using Nemerle.Utility;

namespace Oyster.Units
{
	/// Contains different utilitary fields/methods
	module Util
	{
		// Some constants
		public UnitsNS = "Oyster.Units";

		public UnitInterfacePrefix = "IUnit_";
		public UnitInterfaceFullPrefix : string = $"$UnitsNS.$UnitInterfacePrefix";

		public MacrosSuffix = "Macros";
		public MacrosNS : string = $"$UnitsNS.$MacrosSuffix";

		// Standard environments
		public UnitsEnv : GlobalEnv = GetUnitsEnv("");
		public MacrosEnv : GlobalEnv = GetUnitsEnv(MacrosSuffix);


		/// Creates new GlobalEnv for units global namespace
		public GetUnitsEnv(nsName : string) : GlobalEnv
		{
			def nsName = if (nsName == "") "" else $".$nsName";
			GlobalEnv($"$UnitsNS$nsName&$UnitsNS&&")
		}

		/// Generates PExpr for dot-separated names
		public UseSiteExpr(name : string) : PExpr
		{
			def prepareExpr(l)
			{
				| h :: [] => <[ $(h : usesite) ]>
				| h :: t => <[ $(prepareExpr(t)).$(h : usesite) ]>
				| _ => Message.FatalError($"Expecting dot-separated class name, got $name")
			};
			prepareExpr(GetTypeNameList(name).Reverse())
		}

		/// Returns namespace path list for given full dot-separated type name
		public GetTypeNameList(name : string) : list[string]
		{
			NString.Split(name, '.')
		}

		/// Generates unit interface suffix basing on base unit powers
		public GetUnitSuffix(powers : list[int]) : string
		{
			powers.ToString("_")
		}

		/// Parses unit interface suffix
		public ParseUnitSuffix(t : TypeInfo) : list[int]
		{
			def isIfaceName(s) { s.StartsWith(UnitInterfaceFullPrefix) };
			def name =
				if (t is TypeBuilder)
				{
					// Get for newly generated type
					match (t.GetSuperTypes().Find( fun(super) { isIfaceName(super.TypeInfo.FullName) }))
					{
						| Some(iface) => iface.TypeInfo.Name
						| _ => ""
					}
				}
				else
				{
					// Get for existing type
					def iface = Array.Find(t.SystemType.GetInterfaces(), fun (i) { isIfaceName(i.FullName) });
					if (iface != null) iface.Name else ""
				};
			if (name == "") [] else NString.Split(name, '_').Tail.Map(int.Parse)
		}

		/// Returns PExpr type
		public GetExprType(ctx : Typer, expr : PExpr) : option[TypeInfo]
		{
			match (ctx.TypeExpr(expr).ty.Hint)
			{
				| Some(t) => Some(t.TypeInfo)
				| _ => None()
			}
		}

		/// Returns normalization coefficient if type is concrete unit type
		public GetUnitNormCoef(t : TypeInfo) : option[double]
		{
			| t is TypeBuilder =>
				// Get attribute using Nemerle approach
				match (t.FindAttribute(GlobalEnv.Core.LookupType(GetTypeNameList($"$UnitsNS.NormCoefAttribute")).Value))
				{
					| Some(<[ NormCoef($(coef : double)) ]>) => Some(coef)
					| _ => None()
				}
			| _ =>
				// "Usual" attribute gathering
				match (Attribute.GetCustomAttribute(t.SystemType, typeof(NormCoefAttribute)))
				{
					| null => None()
					| attr => Some((attr :> NormCoefAttribute).NormCoef)
				}
		}

		// Different expression generators
		public AddExpr(lexpr : PExpr, rexpr : PExpr) : PExpr
		{
			<[ $lexpr + $rexpr ]>
		}
		public SubExpr(lexpr : PExpr, rexpr : PExpr) : PExpr
		{
			<[ $lexpr - $rexpr ]>
		}
		public MulExpr(lexpr : PExpr, rexpr : PExpr) : PExpr
		{
			<[ $lexpr * $rexpr ]>
		}
		public DivExpr(lexpr : PExpr, rexpr : PExpr) : PExpr
		{
			<[ $lexpr / $rexpr ]>
		}
	}
}
